//    amazing -- A first-person maze game for the TI-89
//    (C) 2003 by Jeremy Stanley
//
//		This program is free software; you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation; either version 2 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program; if not, write to the Free Software
//    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA


#define USE_TI89              // Produce .89z File

#define OPTIMIZE_ROM_CALLS    // Use ROM Call Optimization
#define NO_EXIT_SUPPORT				// since it isn't needed

// screen is saved manually--not by TIGCC

#include <tigcclib.h>

//#define FRAMECOUNTER 


#include "ang.h"
#include "world.h"
#include "graphics.h"

#define WALLD (FIX_1/3)

#ifdef TEXTURE_WALLS
#define MSTEP (FIX_1/6)
#define ASTEP (ANG_360 / 32)	// = 18 units or 11.25 degrees
char *szEndMsg = "Amazing-89 v2.0 by Jeremy Stanley";
#else
#define MSTEP (FIX_1/8)
#define ASTEP (ANG_360 / 48)	// = 12 units or 7.5 degrees
char *szEndMsg = "Amazing-89 v2.0-nt by Jeremy Stanley";
#endif	// TEXTURE_WALLS

#define MAPW 27
#define MAPH 27
#define MAPSIZE (MAPW * MAPH)

char *szWaitMsg = "Working...";
char *szOutOfMemory = "Insufficient memory";

void drawscreen(void *pBuf);
int gameloop(unsigned char *buf);
void winmessage();

void _main(void)
{
	unsigned char *buf;	        // buffer for video, lcd, maze map, and trig tables
	unsigned char *lcd, *map, *trig_t;
	int res;
	INT_HANDLER int1, int5;

	// allocate memory
	// [I do this all in one malloc due to the paucity of memory handles on the 89
	//  and to simplify error management (only one test)]
  buf = malloc(BUFSIZE + BUFSIZE + TOTSIZE + MAPSIZE + 3840);
  if(!buf) {
  	ST_showHelp(szOutOfMemory);
  	return;
  }
  
	// setup pointers to other things
  lcd = buf + BUFSIZE + BUFSIZE;	// reserve space for light and dark planes
	trig_t = lcd + 3840;  
	map = trig_t + TOTSIZE; 
	
#ifndef PRECOMPUTED_TABLES
	ST_showHelp(szWaitMsg);
	ang_initialize(trig_t);
#endif

  // seed the random number generator
#ifdef FRAMECOUNTER
	srand(1);	// for benchmarking various options, I want the same maze each time
#else
	randomize();
#endif

	// generate a random map
	world_setrandmap(map, MAPW, MAPH, 1);

	// save the old screen 
	LCD_save(lcd);
	   
#ifdef FRAMECOUNTER
	if(!OSRegisterTimer(USER_TIMER, 20))
	{
	   free(buf);
	   return;
	}
#else
	int1 = GetIntVec(AUTO_INT_1);
	int5 = GetIntVec(AUTO_INT_5);
  SetIntVec(AUTO_INT_1, DUMMY_HANDLER);
  SetIntVec(AUTO_INT_5, DUMMY_HANDLER);
#endif

	// enable grayscale mode
  if(!GrayOn())
  {
  	ST_showHelp(szOutOfMemory);
  	free(buf);
  	return;
  }
  // all TIOS graphics functions will draw to the dark plane by default
  PortSet(GetPlane(DARK_PLANE), 239, 127);
  FontSetSys(F_6x8);

	// clear the screen
	memset(GetPlane(LIGHT_PLANE), 0, LCD_SIZE);
	memset(GetPlane(DARK_PLANE), 0, LCD_SIZE);

  // draw the map and border
  pixelcol89(96, 0, 96, GetPlane(DARK_PLANE), 30);
	world_drawmap(128 - MAPW, 50 - MAPH);

	// enter the game loop
  res = gameloop(buf);

	// show the "win message" if appropriate
	if(res)
	   winmessage();

	// reset graphics output and turn off grayscale
	PortRestore();	
	GrayOff();

	// restore interrupt vectors
#ifdef FRAMECOUNTER
  OSFreeTimer(USER_TIMER);
#else
  SetIntVec(AUTO_INT_1, int1);
  SetIntVec(AUTO_INT_5, int5);
#endif

  // restore the old screen
  LCD_restore(lcd);

	// display end message
  ST_showHelp(szEndMsg);
	
	// release all allocated memory
	free(buf);
	return;
}

// returns 1 if the user completed the maze
//  0 otherwise
int gameloop(unsigned char *buf)
{	
  // utility variables
  short u, v, nx, ny, side;
  FIX ix, iy, b;
  unsigned char c, keys;
#ifdef FRAMECOUNTER
  int frames = 0;
#endif

  // player state
  FIX x, y, dx, dy;
  short angle;
  
  // find a random open square
  u = rand() % m_width;
  v = rand() % m_height;
  while(get(u, v) != 0)
  {
    u++; if(u >= m_width) u = 0;
    v++; if(v >= m_height) v = 0;
  }
  
  // initialize player's position and orientation
  x = INT_TO_FIX(u) + FIX_1/2;
  y = INT_TO_FIX(v) + FIX_1/2;
  angle = ANG_90 * (rand() & 3);
  
	// test spin
	/*
	for(i = 0; i < ANG_360; i += 8)
	{ 
		memset(buf, 0, BUFSIZE);
		world_render3D(x, y, i, buf, BUFPITCH);
		drawscreen(buf);
	}
	*/
	
	// Enter main game loop
  while(1)
  {
	// ... compute dx and dy based on player's input ...
	// ... increment or decrement angle based on player's input ...

    // Check for ESC; exit immediately if pressed
    keys = _rowread(~64);
    if(keys & 1) return 0;
    
    // Check for directional movement
    dx = 0; dy = 0;
    keys = _rowread(~1);
    
	  // forward or backward motion
    if(keys & 5)  { 
       dx = FIX_MUL(MSTEP, cos_f(angle));
       dy = -FIX_MUL(MSTEP, sin_f(angle));
       if(keys & 4) { dx = -dx; dy = -dy; } // backward
    }
    
    // sideways motion
    if(keys & 48) {
       side = angle - ANG_90;
       if(side < 0) side += ANG_360;
       if(keys & 32) {
         dx += FIX_MUL(MSTEP, cos_f(side));
         dy -= FIX_MUL(MSTEP, sin_f(side));
       }
       else {
         dx -= FIX_MUL(MSTEP, cos_f(side));
         dy += FIX_MUL(MSTEP, sin_f(side));
       }
    }

    // Check for rotational motion
    if(keys & 2) { // left
       angle += ASTEP;
       if(angle >= ANG_360) angle -= ANG_360;
    }
    if(keys & 8) { // right
       angle -= ASTEP;
       if(angle < 0) angle += ANG_360;
    }

    // move the player
		ix = FIX_IPART(x); nx = FIX_TO_INT(x);
		iy = FIX_IPART(y); ny = FIX_TO_INT(y);
	
	  x += dx; 
    y += dy;
    b = -1;
	
	  // collision detection - x direction
	  if(dx > 0) {
	    if(0 != (c = get(nx+1, ny))) {
	      b = (FIX_1 - WALLD) + ix;
	      if(x > b) {
	        if(c==2) return 1;
	        else x = b;
	      }
	    }
	  }
    else
    {
      if(0 != (c = get(nx-1, ny))) {
        b = WALLD + ix;
        if(x < b)  {
          if(c==2) return 1;
          else x = b;
        }
      }
    }

    // collision detection - y direction
    if(dy > 0) {
       if(0 != (c = get(nx, ny+1))) {
          b = (FIX_1 - WALLD) + iy;
          if(y > b) {
              if(c==2) return 1;
              else y = b;
          }
       }
    }
    else
    {
      if(0 != (c = get(nx, ny-1))) {
        b = WALLD + iy;
        if(y < b) {
            if(c==2) return 1;
            else y = b;
        }
      }
    }

    // The two above cases will not handle walking straight into a jutting corner
    if(b == -1)
    {
	    if(get(FIX_TO_INT(x + dx), FIX_TO_INT(y + dy)))
	    {
	      x -= dx + dx;
	      y -= dy + dy;
	    }
    }

    // Finally, draw the maze

#ifdef TEXTURE_WALLS
		// draw light plane background
		memset(buf, 0, BUFSIZE/2);
		memset(buf+BUFSIZE/2, 0xFF, BUFSIZE/2);
		// draw dark plane background
		memset(buf + BUFSIZE, 0, BUFSIZE);
#else
    memset(buf, 0, 2*BUFSIZE);
#endif

		world_render3D(x, y, angle, buf);
		drawscreen(buf);
		
#ifdef FRAMECOUNTER
		frames++;
		if(OSTimerExpired(USER_TIMER))
		{
			sprintf(buf, "%03d", frames);
			DrawStr(100, 0, buf, A_REPLACE);
			frames = 0;
		}
#endif
  } // while(1)

}

#if 0

// JS 9/28/03--WTF?  memcpy() is a ROM CALL?  And I've been doing it
// 192 times per frame?  Sheesh, I thought it would be an intrinsic function
// like it is on VC++.  Better ask the TIGCC crew for fast asm versions...
// or maybe I should get off my duff, learn 68k assembly, and do it myself.
// I used to write stuff for Zshell, for crying out loud.  :)

// In the mean time, however, I'll try doing an archaic loop and see how that fares...
// ... see below...

void drawscreen(void *pBuf)
{
	short i;
	unsigned char *pLight = GetPlane(LIGHT_PLANE), *pDark = GetPlane(DARK_PLANE);
//	static unsigned short throbber = 0;

	for(i = 0; i < HEI; i++)
	{
		memcpy(pLight, pBuf, BUFPITCH);
		memcpy(pDark, pBuf + BUFSIZE, BUFPITCH);
		pLight += 30;
		pDark += 30;
		pBuf += BUFPITCH;
	}
	
	/*
	throbber >>= 1;
	if(!throbber) {
		throbber = 0x8000u;
	}
	*((unsigned short *)LCD_MEM + 7) = throbber;
	*/
}

#else

void drawscreen(void *buf)
{
	short i;
	unsigned short *pLight = GetPlane(LIGHT_PLANE), *pDark = GetPlane(DARK_PLANE);
	unsigned short *pBuf = (unsigned short *)buf;

#if (BUFPITCH & 1) != 0
#error bad screen width
#endif

	for(i = 0; i < HEI; i++)
	{
		unsigned short *line_end = pBuf + BUFPITCH/2;
		while(pBuf < line_end)
		{
			*pDark++ = *(pBuf + BUFSIZE/2);
			*pLight++ = *pBuf++;
		}
		pLight += (15 - BUFPITCH/2);
		pDark += (15 - BUFPITCH/2);
	}
}

#endif

void winmessage()
{
   FontSetSys(F_8x10);
   DrawStr(16, 45, "Congratulations!", A_REPLACE);
   while(!(_rowread(~64) & 1));
}
